#include <stdlib.h>
#include <math.h>
#include "geosim.h"
#include "gspriv.h"
#include "windfunc.h"
#include "ifw.h"
#include "rootw.h"
#include "objprop.h"
#include <string.h>
#include "iemisc.h"
#include "infprocs.h"

#define SWAP(a, b, t) t=a;a=b;b=t;
#define TAB_FOR_SS

typedef enum {dNothing, dLine, dRect, dGrowRect} DragObjType;

/* This would have been *so* much easier had I used C++ ... */

/* from iefile.c: */
extern GSColor IEInterface_Palette[];
extern char *IEColorlist[];
extern char *IEAlias[], *IEAcolor[];
boolean showPropCreate;
/* from GIL:*/
extern char Interface_Palette[];
extern char *Colorlist[];
extern char *Alias[], *Acolor[];
extern Iconfig InfState;

extern boolean interfaceWind;
extern IEIconfig iestate;

/* internal protos */					
void SetDefaultObjects();
void ClickObject(int x, int y, DragStatusType status);
void StartToolDrag(int x, int y);
void EndToolDrag();
void GetCurObjectDims(int *, int *, int *, int *);
void SetCurObjectLoc(int, int);
void DrawIEField(IEField *fld, int wx, int wy);
void DrawIEDragArea(IEDragArea *dr, int wx, int wy);
void StripArrowSym(char *btext, char *c, IEButton *b);
void DrawIEButton(IEButton *, int, int);
void DrawIEGraphic(IEGraphic *, int, int);
void DrawIEPopup(IEPopup *, int, int, int, int);
void DrawInterface(boolean);
void UpdateStateButtons(int);
/*void SwapPalette(
	char oldPalette[], char newPalette[],
	char *oldCollist[], char *newCollist[],
	char *oldAlias[], char *newAlias[],
	char *oldAColor[], char *newAColor[],
	char *oldbTL, char *newbTL,
	char *oldbBR, char *newbBR
	);*/
	
void SelectObjectHitBy(int, int, DragStatusType);
void DragItem(int, int, DragStatusType);
void ReleaseItem(int, int, DragStatusType);
void StartCreateObject(int, int, DragStatusType);
void GrowItem(int, int, DragStatusType);
void EndCreateObject(int, int, DragStatusType);
void ToggleToolWind();
boolean ToolWindowEvent(int, int, DragStatusType);

/* Internal global variables - let's keep these to a minimum. */

	/* drag stuff */
static int offx, offy;  /* offset */
static int dragx, dragy, dragwd, draght; /* drag box dims */
static int dbx, dby, dbwd, dbht; /* drag bounds */
static DragObjType curDrag;

	/* current object info */
static IECurObj cobject;

	/* tool window stuff */
static int twindx, twindy, twindwd, twindht;
/*static int tdragx, tdragy, tdragwd, tdraght;*/
boolean twindVisible = FALSE;

	/* state stuff */
#define SEL_STATE 0
#define NEW_WINDOW 1
#define NEW_BUTTON 2
#define NEW_RECT 3
#define NEW_LINE 4
#define NEW_LABEL 5
#define NEW_FIELD 6
#define NEW_DRAGAREA 7
#define NEW_POPUP 8
#define NUM_STATES 9

char *StateButtonName[NUM_STATES+1] = 
	{"SelState", "WindState", "BtnState", "RectState", 
	"LineState", "LabelState", "FieldState", "DragAreaState", 
	"PopupState", NULL};
	
int curState;


/* Default object information */
IEWindow dfltWindow; /* newly created window */
IEButton dfltButton; /* newly created button */
IEDragArea dfltDragArea;
IEField dfltField;
IEPopup dfltPopup;
static int numWinds = 0, numButtons = 0, 
		numDragAreas = 0, numFields = 0, numPopups = 0;
void (*dragProcessFunc)(int, int, DragStatusType);

/**/

void StartToolDrag(int x, int y)
{
	dragx = twindx;
	dragy = twindy;
	dragwd = twindwd;
	draght = twindht;

	offx = x - twindx;
	offy = y - twindy;

	dbx = 0;
	dby = 0;
	dbwd = 640;
	dbht = 480;

	curDrag = dRect;

	GSdrawxorbox(dragx, dragy, dragwd, draght);
}

void EndToolDrag()
{
	GSdrawxorbox(dragx, dragy, dragwd, draght);
	GSremovenamedwindow("toolw");
	
	GSsetwindowspecs("toolw", &dragx, &dragy, NULL, NULL);

	twindx = dragx;
	twindy = dragy;
	
	GSdrawnamedwindow("toolw", BSTORE, NO_POPUP);
}
		

void GetCurObjectDims(int *x, int *y, int *wd, int *ht)
{
/* the size of this function is ridiculous - fix it */
	if (!cobject.cwind) {
		*x = *y = *wd = *ht = 0;
		return;
		}
	
	*x = cobject.cwind->x;
	*y = cobject.cwind->y;
	
	if (cobject.objtype == iewind) {
		*wd = cobject.cwind->wd;
		*ht = cobject.cwind->ht;
		return;
		}
		
	switch (cobject.objtype) {
		case iebutton:
			*x += cobject.cbutton->x;
			*y += cobject.cbutton->y;
			*wd = cobject.cbutton->wd;
			*ht = cobject.cbutton->ht;
			break;
		case iedragarea:
			*x += cobject.cdrag->x;
			*y += cobject.cdrag->y;
			*wd = cobject.cdrag->wd;
			*ht = cobject.cdrag->ht;
			break;
		case iegraphic:
			if ((cobject.cgraphic->gr.type == GR_LINE) ||
				(cobject.cgraphic->gr.type == GR_WLINE)) {
					*x += MIN(cobject.cgraphic->gr.x, cobject.cgraphic->gr.wd);
					*wd = cobject.cgraphic->gr.x - cobject.cgraphic->gr.wd;
					*wd = MAX(*wd, -(*wd));
					
					*y += MIN(cobject.cgraphic->gr.y, cobject.cgraphic->gr.ht);
					*ht = cobject.cgraphic->gr.y - cobject.cgraphic->gr.ht;
					*ht = MAX(*ht, -(*ht));
					}
			else {
				*x += cobject.cgraphic->gr.x;
				*y += cobject.cgraphic->gr.y;
				*wd = cobject.cgraphic->gr.wd;
				*ht = cobject.cgraphic->gr.ht;
				}
			break;
		case iefield:
			*x += cobject.cfield->x;
			*y += cobject.cfield->y;
			*wd = cobject.cfield->wd;
			*ht =  cobject.cfield->ht;
			break;
		}
}

void SetCurObjectLoc(int x, int y)
{
	/* x and y are given in "global" coordinates */
	
	if (cobject.objtype == iewind) {
		cobject.cwind->x = x;
		cobject.cwind->y = y;
		return;
		}
	if (cobject.objtype == iepopup) {
		cobject.cpopup->x = x;
		cobject.cpopup->y = y;
		return;
		}
		
	x -= cobject.cwind->x;
	y -= cobject.cwind->y;
	
	switch (cobject.objtype) {
		case iebutton:
			cobject.cbutton->x = x;
			cobject.cbutton->y = y;
			break;
		case iedragarea:
			cobject.cdrag->x = x;
			cobject.cdrag->y = y;
			break;
		case iegraphic:
			cobject.cgraphic->gr.x = x;
			cobject.cgraphic->gr.y = y;
			if ((cobject.cgraphic->gr.type == GR_LINE) ||
				(cobject.cgraphic->gr.type == GR_WLINE)) {
					/* cheat here: use dragwd, draght */
					cobject.cgraphic->gr.wd = x + dragwd;
					cobject.cgraphic->gr.ht = y + draght;
					}
			break;
		case iefield:
			cobject.cfield->x = x;
			cobject.cfield->y = y;
			break;
		}
}

void ClickObject(int x, int y, DragStatusType status)
{
	SelectObjectHitBy(x, y, status);
	
	if (curDrag == dNothing)
		return;
		
	switch (cobject.objtype) {
	
		case iewind:
			DrawWPropWind(cobject.cwind, FALSE);
			return;
	
		case iepopup:
			PopupToProp(cobject.cpopup);
			break;
			
		case iebutton:
			ButtonToProp(cobject.cbutton, cobject.cwind);
			break;
			
		case iedragarea:
			DragAreaToProp(cobject.cdrag, cobject.cwind);
			break;
			
		case iegraphic:
			GraphicToProp(cobject.cgraphic, cobject.cwind);
			break;
				
		case iefield:
			FieldToProp(cobject.cfield, cobject.cwind);
			break;
					
		default:
			return;
		}

	DrawPropWind(FALSE);
}

void SelectObjectHitBy(int x, int y, DragStatusType status)
{
	IEWindow *wind;
	IEPopup *popup;
	IEButton *bt=NULL;
	IEGraphic *gr=NULL;
	IEField *fld=NULL;
	IEDragArea *da=NULL;
	int localx, localy;
	/* find window first */
	
	wind = iestate.winlisttail;
	popup = iestate.popuplisttail;
	curDrag = dNothing;
	
	while (popup != NULL) {
		if (popup->visibleineditor &&
			point_in_rect(x, y, popup->x,
					popup->y, popup->wd, popup->ht)) {
					cobject.objtype = iepopup;
					cobject.cpopup = popup;
					dragx = popup->x;
					dragy = popup->y;
					dragwd = popup->wd;
					draght = popup->ht;
					dbx = 0;
					dby = 0;
					dbwd = 640;
					dbht = 480;
					curDrag = dRect;
					goto exit;
					}
		popup = popup->prevpopup;
		}
	
	while (wind != NULL) {
		if (wind->visibleineditor &&
			point_in_rect(x, y, wind->x, 
					wind->y, wind->wd, wind->ht))
			break;
		else
			wind = wind->prevwin;
		}
	
	cobject.cwind = wind;
	
	if (wind == NULL)
		return;
	
	dragProcessFunc = GrowItem;

	localx = x - wind->x;
	localy = y - wind->y;

	/* now find item hit */

	bt = wind->lastbutton;
	
	while (bt != NULL) {
		if (bt->visibleineditor && 
			point_in_rect(localx, localy, bt->x, 
						bt->y, bt->wd, bt->ht))
			break;
		else
			bt = bt->prevbutton;
		}	
	if (bt==NULL) {
		da = wind->lastdragarea;
		while (da != NULL) {
			if (point_in_rect(localx, localy,
				da->x, da->y, da->wd, da->ht))
				break;
			else
				da = da->prevdragarea;
			}
		}
	if (bt==NULL && da==NULL) {
		fld = wind->lastfield;
		while (fld != NULL) {
			if (/*fld->visibleineditor &&*/
				point_in_rect(localx, localy, 
					fld->x, fld->y, fld->wd, fld->ht))
				break;
			else
				fld = fld->prevfield;
			}
		}
	if (bt==NULL && da==NULL && fld==NULL) {
		gr = wind->lastgraphic;
		while (gr != NULL) {
			if ((gr->gr.type == GR_LINE) ||
					(gr->gr.type == GR_WLINE)) {
					if (point_hits_line(localx, localy, 
						gr->gr.x, gr->gr.y, gr->gr.wd, gr->gr.ht))
							break;
					}
			else if (point_in_rect(localx, localy,
					gr->gr.x, gr->gr.y, gr->gr.wd, gr->gr.ht))
						break;
						
			gr = gr->prevgraphic;
			}
		}
	if (bt) {
		cobject.objtype = iebutton;
		cobject.cbutton = bt;
		dragx = bt->x + wind->x;
		dragy = bt->y + wind->y;
		dragwd = bt->wd;
		draght = bt->ht;
		dbx = wind->x;
		dby = wind->y;
		dbwd = wind->wd;
		dbht = wind->ht;
		curDrag = dRect;
		}
	else if (gr) {
		cobject.objtype = iegraphic;
		cobject.cgraphic = gr;
		dragx = gr->gr.x + wind->x;
		dragy = gr->gr.y + wind->y;
		dbx = wind->x;
		dby = wind->y;
		dbwd = wind->wd;
		dbht = wind->ht;
		if ((gr->gr.type == GR_LINE) || (gr->gr.type == GR_WLINE)) {
			dragwd = gr->gr.wd - gr->gr.x;
			draght = gr->gr.ht - gr->gr.y;
			curDrag = dLine;
			}
		else {
			dragwd = gr->gr.wd;
			draght = gr->gr.ht;
			curDrag = dRect;
			}
		}
	else if (fld) {
		cobject.objtype = iefield;
		cobject.cfield = fld;
		dragx = fld->x + wind->x;
		dragy = fld->y + wind->y;
		dragwd = fld->wd;
		draght = fld->ht;
		dbx = wind->x;
		dby = wind->y;
		dbwd = wind->wd;
		dbht = wind->ht;
		curDrag = dRect;
		}
	else if (da) {
		cobject.objtype = iedragarea;
		cobject.cdrag = da;
		dragx = da->x + wind->x;
		dragy = da->y + wind->y;
		dragwd = da->wd;
		draght = da->ht;
		dbx = wind->x;
		dby = wind->y;
		dbwd = wind->wd;
		dbht = wind->ht;
		curDrag = dRect;
		}	
	else {
		cobject.objtype = iewind;
		dragx = wind->x;
		dragy = wind->y;
		dragwd = wind->wd;
		draght = wind->ht;
		dbx = 0;
		dby = 0;
		dbwd = 640;
		dbht = 480;
		curDrag = dRect;
		}
exit:
	offx = x - dragx;
	offy = y - dragy;
	
	GSsetlinesize(1);
	
	if (status == DRAG_INIT) {
		if (curDrag == dLine)
			GSdrawxorline(dragx, dragy, dragx+dragwd, dragy+draght);
		else if (curDrag == dRect)
			GSdrawxorbox(dragx, dragy, dragwd, draght);
		dragProcessFunc = DragItem;
		}
}

void StartCreateObject(int x, int y, DragStatusType status)
{
	IEWindow *wind = iestate.winlisttail;
	curDrag = dNothing;
	
	dragProcessFunc = GrowItem; /* default - labels use DragItem */

	if (curState == NEW_WINDOW) {
		
		dbx = 0;
		dby = 0;
		dbwd = 640;
		dbht = 480;
		
		dragx = dragwd = x;
		dragy = draght = y;
		
		dragProcessFunc = GrowItem;
		curDrag = dGrowRect;
		}
	else if (curState == NEW_POPUP) {
		
		dbx = 0;
		dby = 0;
		dbwd = 640;
		dbht = 480;
		
		offx = 0;
		offy = 0;
		
		dragx = x;
		dragy = y;
		dragwd = 60;
		draght = LARGE;
		dragProcessFunc = DragItem;
		curDrag = dRect;
		}	
		
	else {
		while (wind != NULL) {
			if (wind->visibleineditor &&
				point_in_rect(x, y, wind->x, 
						wind->y, wind->wd, wind->ht))
				break;
			else
				wind = wind->prevwin;
			}
		
		if (wind == NULL)
			return;
		else
			cobject.cwind = wind;
		
		dragProcessFunc = GrowItem;
		dbx = wind->x;
		dby = wind->y;
		dbwd = dbx + wind->wd;
		dbht = dby + wind->ht;
		
		dragx = dragwd = x;
		dragy = draght = y;

		if (curState == NEW_LINE)
			curDrag = dLine;
		else if (curState == NEW_LABEL) {
			curDrag = dRect;
			dragx = x;
			dragy = y - LARGE;
			dragwd = 30;
			draght = LARGE;
			offx = 0;
			offy = LARGE;
			dbwd = wind->wd;
			dbht = wind->ht;
			dragProcessFunc = DragItem;
			}
		else
			curDrag = dGrowRect;
		}
	
/*	dragx = dragwd;
	dragy = draght;*/
	
	GSsetlinesize(1);

	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragwd, draght);
	else if (curDrag == dGrowRect)
		GSdrawxorbox(MIN(dragx, dragwd), MIN(dragy, draght),
							abs(dragwd-dragx), abs(draght-dragy));
	else if (curDrag == dRect)
		GSdrawxorbox(dragx, dragy, dragwd, draght);
}

/* DragItem :
	If you're dragging an object, dragging is clipped to the
	object's window.
	If you're dragging a window, dragging is clipped to the 
	screen.
	
	Depends on globals dragx, dragy, dragwd, draght,
						dbx, dby, dbwd, dbht,
						curDrag
*/

void DragItem(int x, int y, DragStatusType status)
{
	
/* erase old xor box */
	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragx+dragwd, dragy+draght);
	else if (curDrag == dRect)
		GSdrawxorbox(dragx, dragy, dragwd, draght);
	else 
		return;
		
	dragx = x-offx;
	dragy = y-offy;

	if (dragx < dbx)
		dragx = dbx;
	else if (dragx + dragwd > dbx + dbwd)
		dragx = dbx + dbwd - dragwd;
	
	if (dragy < dby)
		dragy = dby;
	else if (dragy + draght > dby + dbht)
		dragy = dby + dbht - draght;
		
	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragx+dragwd, dragy+draght);
	else if (curDrag == dRect)
		GSdrawxorbox(dragx, dragy, dragwd, draght);
}


void GrowItem(int x, int y, DragStatusType status)
{
	/* dragx, dragy are current corner;  dragwd, draght orig corner */
	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragwd, draght);
	else if (curDrag == dGrowRect)
		GSdrawxorbox(MIN(dragx, dragwd), MIN(dragy, draght),
							abs(dragwd-dragx), abs(draght-dragy));
	else 
		return;
		
		
	if (x < dbx)
		x = dbx;
	else if (x > dbwd)
		x = dbwd;
		
	if (y < dby)
		y = dby;
	else if (y > dbht)
		y = dbht;
	
	dragx = x;
	dragy = y;
	
	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragwd, draght);
	else if (curDrag == dGrowRect)
		GSdrawxorbox(MIN(dragx, dragwd), MIN(dragy, draght),
							abs(dragwd-dragx), abs(draght-dragy));
}

void ReleaseItem(int x, int y, DragStatusType status)
{
	int oldx, oldy, oldwd, oldht;
	int clipx, clipy, clipwd, clipht;
	boolean replaceTWind = FALSE;
	IEWindow *w = cobject.cwind;
	
	dragProcessFunc = NULL;
	
	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragx+dragwd, dragy+draght);
	else if (curDrag == dRect)
		GSdrawxorbox(dragx, dragy, dragwd, draght);
	else 
		return;
		
	if (cobject.objtype == iewind || cobject.objtype == iepopup) {
	
		if (cobject.objtype == iewind) {
			oldx = w->x;
			oldy = w->y;
			oldwd = w->wd;
			oldht = w->ht;
			}
		else {
			oldx = cobject.cpopup->x;
			oldy = cobject.cpopup->y;
			oldwd = cobject.cpopup->wd;
			oldht = cobject.cpopup->ht;
			}
			
		calc_union_rect(oldx-1, oldy-1, oldwd+2, oldht+2,
						dragx-1, dragy-1, dragwd+2, draght+2,
						&clipx, &clipy, &clipwd, &clipht);
		SetCurObjectLoc(dragx, dragy);
		if (r_intersect(clipx, clipy, clipwd, clipht,
						twindx, twindy, twindwd, twindht)) {
				replaceTWind = TRUE;
				ToggleToolWind();
				} 
		PushClip(clipx, clipy, clipwd, clipht);
		w = iestate.winlisthead;
		DrawFromWind(w, clipx, clipy, clipwd, clipht);
		ResetClip();
		}
	else {		
		GetCurObjectDims(&oldx, &oldy, &oldwd, &oldht);
		SetCurObjectLoc(dragx, dragy);
		if (dragwd < 0) {
			dragx += dragwd;
			dragwd = -dragwd;
			}
		if (draght < 0) {
			dragy += draght;
			draght = -draght;
			}

		calc_union_rect(oldx-1, oldy-1, oldwd+2, oldht+2,
						dragx-1, dragy-1, dragwd+2, draght+2,
						&clipx, &clipy, &clipwd, &clipht);

		if (r_intersect(clipx, clipy, clipwd, clipht,
						twindx, twindy, twindwd, twindht)) {
				replaceTWind = TRUE;
				ToggleToolWind();
				} 
		
		/*GSPset_clip_rect(clipx, clipy, clipwd, clipht);*/
		PushClip(clipx, clipy, clipwd, clipht);

		DrawFromWind(w, clipx, clipy, clipwd, clipht);
		
		/*GSPset_clip_rect(0, 0, 640, 480);*/
		ResetClip();
		}
		
	if (replaceTWind)
		ToggleToolWind();
}

void EndCreateObject(int x, int y, DragStatusType status)
{	
	IEWindow *w = NULL, *ww;
	IEButton *bt;
	IEGraphic *gr;
	IEField *fld;
	IEDragArea *dr;
	IEPopup *newpop = NULL;
	
	int dclip_x, dclip_y, dclip_wd, dclip_ht;
	
	dragProcessFunc = NULL;
	
	if (curDrag == dLine)
		GSdrawxorline(dragx, dragy, dragwd, draght);
	else if (curDrag == dGrowRect)
		GSdrawxorbox(MIN(dragx, dragwd), MIN(dragy, draght),
							abs(dragwd-dragx), abs(draght-dragy));
	else if (curDrag == dRect)
		GSdrawxorbox(dragx, dragy, dragwd, draght);
	else 
		return;
		
	dclip_x = MIN(dragx, dragwd);
	dclip_y = MIN(dragy, draght);
	dclip_wd = abs(dragwd-dragx);
	dclip_ht = abs(draght-dragy);

	if (dragx == dragwd && draght == dragy)
		return;
		
	if (curState == NEW_WINDOW) {
		
		w = (IEWindow *) malloc(sizeof(IEWindow));
		if (w == NULL)
			GSelog(-1, "Out of memory for new window.\n");
			
		memcpy(w, &dfltWindow, sizeof(IEWindow));
		
		w->x = MIN(dragx, dragwd);
		w->y = MIN(dragy, draght);
		w->wd = abs(dragwd-dragx);
		w->ht = abs(draght-dragy);
		
		sprintf(w->name, "NewWindow%d", ++numWinds);
			
       	if (iestate.winlisthead == NULL)
       		iestate.winlisthead = w;
       	else
       		iestate.winlisttail->nextwin = w;
       	w->prevwin = iestate.winlisttail;
       	iestate.winlisttail = w;
       	w->nextwin = NULL;
       	ww = w;
   		}
   	else if (curState == NEW_POPUP) {
   		
   		newpop = (IEPopup *) malloc(sizeof(IEPopup));
   		if (newpop == NULL)
   			GSelog(-1, "Out of memory for new popup.\n");
   		memcpy(newpop, &dfltPopup, sizeof(IEPopup));
   		sprintf(newpop->title, "NewPopup%d", ++numPopups);
   		dclip_x = newpop->x = dragx;
   		dclip_y = newpop->y = dragy;
   		dclip_wd = newpop->wd = dragwd;
   		dclip_ht = newpop->ht = draght;
   		if (iestate.popuplisthead == NULL)
   			iestate.popuplisthead = newpop;
   		else
   			iestate.popuplisttail->nextpopup = newpop;
   		newpop->prevpopup = iestate.popuplisttail;
   		iestate.popuplisttail = newpop;
   		newpop->nextpopup = NULL;
   		}
	else {
		w = cobject.cwind;
		if (curState == NEW_FIELD) {
			fld = (IEField *) malloc(sizeof(IEField));
			memcpy(fld, &dfltField, sizeof(IEField));

			fld->x = MIN(dragx, dragwd) - w->x;
			fld->y = MIN(dragy, draght) - w->y;
			fld->wd = abs(dragwd-dragx);
			fld->ht = abs(draght-dragy);
			
			sprintf(fld->name, "NewField%d", ++numFields);
			
			if (w->firstfield == NULL)
				w->firstfield = fld;
			else
				w->lastfield->nextfield = fld;
			fld->prevfield = w->lastfield;
			w->lastfield = fld;
			fld->nextfield = NULL;
			}
		else if (curState == NEW_DRAGAREA) {
			dr = (IEDragArea *) malloc(sizeof(IEDragArea));
			memcpy(dr, &dfltDragArea, sizeof(IEDragArea));
			
			dr->x = MIN(dragx, dragwd) - w->x;
			dr->y = MIN(dragy, draght) - w->y;
			dr->wd = abs(dragwd-dragx);
			dr->ht = abs(draght-dragy);
			dr->active = TRUE;
			sprintf(dr->name, "NewDragArea%d", ++numDragAreas);
			sprintf(dr->funcname, "funcname");
			
			if (w->firstdragarea == NULL)
				w->firstdragarea = dr;
			else
				w->lastdragarea->nextdragarea = dr;
			dr->prevdragarea = w->lastdragarea;
			w->lastdragarea = dr;
			dr->nextdragarea = NULL;
			}
		else if (curState == NEW_BUTTON) {	
			bt = (IEButton *) malloc(sizeof(IEButton));
			
			if (bt == NULL)
				GSelog(-1, "Out of memory for new button.\n");	
			
			memcpy(bt, &dfltButton, sizeof(IEButton));
			
			bt->x = MIN(dragx, dragwd) - w->x;
			bt->y = MIN(dragy, draght) - w->y;
			bt->wd = abs(dragwd-dragx);
			bt->ht = abs(draght-dragy);
			
			sprintf(bt->name, "NewButton%d", ++numButtons);
			
			if (w->firstbutton == NULL)
				w->firstbutton = bt;
			else
				w->lastbutton->nextbutton = bt;
			bt->prevbutton = w->lastbutton;
			w->lastbutton = bt;
			bt->nextbutton = NULL;
			}
		else if ((curState == NEW_RECT)
					||(curState == NEW_LINE)
					||(curState == NEW_LABEL)){
			gr = (IEGraphic *) malloc(sizeof(IEGraphic));
			
			if (gr == NULL)
				GSelog(-1, "Out of memory for new button.\n");
			
			/*memcpy(gr, &dfltGraphic, sizeof(IEGraphic));*/
			
			if (curState == NEW_RECT) {
				gr->gr.type = GR_RECT;
				gr->gr.x = MIN(dragx, dragwd) - w->x;
				gr->gr.y = MIN(dragy, draght) - w->y; 
				gr->gr.wd = abs(dragwd-dragx);
				gr->gr.ht = abs(draght-dragy);
				}
			else if (curState == NEW_LINE) {
				gr->gr.type = GR_LINE;
				gr->gr.x = dragx - w->x;
				gr->gr.y = dragy - w->y; 
				gr->gr.wd = dragwd - w->x;
				gr->gr.ht = draght - w->y;
				dclip_x = dragx;
				dclip_y = dragy;
				dclip_wd = dragwd;
				dclip_ht = draght;
				}
			else if (curState == NEW_LABEL) {
				gr->gr.type = GR_LABEL;
				if ((gr->gr.text = malloc(BIGSTRINGLEN+1)) == NULL)
					GSelog(-1, "Out of memory for label text.\n");
				gr->gr.text[0] = '\0';
				/*set_label_dims(gr, MIN(dragx, dragwd) - w->x, MAX(dragy, draght) - w->y);*/
				set_label_dims(gr, dragx - w->x, dragy + draght - w->y);
				}
			strcpy(gr->grcol, "black");
			strcpy(gr->grshadowcol, "NONE");
			
			if (w->firstgraphic == NULL)
				w->firstgraphic = gr;
			else
				w->lastgraphic->nextgraphic = gr;
			gr->prevgraphic = w->lastgraphic;
			w->lastgraphic = gr;
			gr->nextgraphic = NULL;
			}
				
		}
		
/*	GSPset_clip_rect(dclip_x, dclip_y, dclip_wd, dclip_ht);*/
	PushClip(dclip_x, dclip_y, dclip_wd, dclip_ht);
	
	if (w)
		DrawFromWind(w, dclip_x, dclip_y, dclip_wd, dclip_ht);
	else if (newpop)
		DrawFromPopup(newpop, dclip_x, dclip_y, dclip_wd, dclip_ht);
/*	while (w != NULL) {
		DrawIEWindow(w, dclip_x, dclip_y, dclip_wd, dclip_ht);
		w = w->nextwin;
		}*/
	/*GSPset_clip_rect(0, 0, 640, 480);*/
	ResetClip();
		
	if (showPropCreate) {

		if (curState == NEW_WINDOW)
	       	DrawWPropWind(ww, TRUE);
		else {
			if (curState == NEW_POPUP)
	    		PopupToProp(newpop);
			else if (curState == NEW_FIELD)
				FieldToProp(fld, cobject.cwind);
			else if (curState == NEW_DRAGAREA)
				DragAreaToProp(dr, cobject.cwind);
			else if (curState == NEW_BUTTON)
				ButtonToProp(bt, cobject.cwind);
			else if ((curState == NEW_RECT)
					||(curState == NEW_LINE)
						||(curState == NEW_LABEL))
				GraphicToProp(gr, cobject.cwind);
			else 
				return;
			DrawPropWind(TRUE);
			}
		}
}

#define ESC_SYM '\\'

void StripArrowSym(char *btext, char *c, IEButton *b)
{
	for (; *btext != '\0'; btext++) {
		if (*btext == ESC_SYM) {
			btext++;
			if (*btext != '\0')
				b->arrow = (ArrowSymType) *btext;
			}
		else
			*c++ = *btext;
		}
	*c = '\0';
}
	
static void make_triang(int t[], int x1, int y1, int x2, int y2, int x3, int y3)
{
	t[0] = x1; t[1] = y1; t[2] = x2; t[3] = y2; t[4] = x3, t[5] = y3;
}

void DrawIEButton(IEButton *b, int wx, int wy)
{
	int bx = b->x + wx, by = b->y + wy;
	int i;
	int tx, ty, twd, tht;
	int awd = 0, aht = 0, amarg = 0, twd_max;
	
	char *text2;
	
	BigString text1;
	StripArrowSym(b->text, text1, b);
	
	GSsetcolor(GSconvertcolor(b->bgcolor));
	GSPfill_rect(bx, by, b->wd, b->ht);
	
	if (b->border) {
		GSsetcolor(InfState.buttonborder);
		GSPframe_rect(bx, by, b->wd, b->ht);
		GSsetcolor (InfState.buttonTLcolor);
		GSPline (bx+1, by+1, bx+b->wd-2, by+1);
		GSPline (bx+1, by+1, bx+1, b->ht+by-2);
		GSsetcolor (InfState.buttonBRcolor);
		GSPline (bx+2, by+b->ht-2, bx+b->wd-2, by+b->ht-2);
		GSPline (bx+b->wd-2, by+2, bx+b->wd-2, by+b->ht-2);
		GSsetcolor(LIGHTBUTTONBORDER);
		GSPdraw_pixel (bx+b->wd-2, by+1);
		GSsetcolor(DARKBUTTONBORDER);
		GSPdraw_pixel (bx+1, by+b->ht-1);
		}
     
    if (b->active == ENABLED)
    	GSsetcolor(GSconvertcolor(b->efgcolorname));
    else
    	GSsetcolor(GSconvertcolor(b->dfgcolorname));
    
    GSsetfont(b->tsize);

	text2 = strchr(text1, '@');
	tht = GStextheight(text1);
	
	/* Button text should be centered vertically, whether
	there is one line or two. */
	
	if (text2 == NULL)
		ty = wy + b->y + (b->ht - tht)/2 + tht + (b->ht-tht)%2;
	else
		ty = wy + b->y + b->ht/2;
	
	if (text2 != NULL)
		*text2 = '\0';
		
	twd_max = twd = GStextwidth(text1);
	switch(b->textjust) {
		case LEFT:
			tx = wx+b->x + 5;
			break;
		case RIGHT:
			tx = wx+b->x + b->wd - twd - awd - amarg - 6;
			break;
		case CENTER:
			tx = wx+b->x + (b->wd - twd - awd - amarg) /2;
			break;
		default:		/* for debug purposes */
			GSbell();
			break;
		}
	GSPdraw_string(tx, ty, text1);
	ty += tht;
	
	if (text2 != NULL) {
		text2++;
		for (i=0; (text2[i]!='\0')&&(text2[i]!='@'); i++)
			text1[i] = text2[i];
		text1[i] = '\0';
		twd = GStextwidth(text1);
		switch(b->textjust) {
			case LEFT:
				tx = wx+b->x + 5;
				break;
			case RIGHT:
				tx = wx+b->x + b->wd - twd - awd - amarg - 6;
				break;
			case CENTER:
				tx = wx+b->x + (b->wd - twd - awd - amarg) /2;
				break;
			}
		GSPdraw_string(tx, ty, text1);
		}
	
	if (twd > twd_max)
		twd_max = twd;
		
	awd = b->wd - 2*ARROWMARGIN - twd_max;
	aht = b->ht - 2*ARROWMARGIN;
	
	if (b->arrow != NO_ARROW_SYM) {
		int ac[6];
		int aleft, aright, atop, abottom;
		int adim = MIN(awd, aht);
		aleft = wx+b->x+twd_max + ((awd - adim)/2) + ARROWMARGIN;
		atop = wy+b->y + ((aht - adim)/2) + ARROWMARGIN;
		aright = aleft + adim;
		abottom = atop + adim;
		
		switch(b->arrow) {
			case UP_ARROW_SYM:
				make_triang(ac, (aleft+aright)/2, atop, aleft, abottom, aright, abottom);
				break;
			case DOWN_ARROW_SYM:
				make_triang(ac, (aleft+aright)/2, abottom, aleft, atop, aright, atop);
				break;
			case LEFT_ARROW_SYM:
				make_triang(ac, aleft, atop, aleft, abottom, aright, (atop+abottom)/2);
				break;		
			case RIGHT_ARROW_SYM:
				make_triang(ac, aleft, (atop+abottom)/2, aright, atop, aright, abottom);
				break;
			default:
				return;
			}
		GSPfillpoly(3, ac);
		}
}

void DrawIEGraphic(IEGraphic *g, int wx, int wy)
/* wx and wy should be in "global" coordinates. */
{
	Graphic *gr = &(g->gr);
	gr->color = GSconvertcolor(g->grcol);
	gr->shadowcolor = GSconvertcolor(g->grshadowcol);
	GSsetlinesize(THIN);
	
	switch (gr->type) {
	
		case GR_LINE:
			GSsetcolor(gr->color);
			GSPline(wx+gr->x, wy+gr->y,
				wx+gr->wd, wy+gr->ht);
			break;
		
		case GR_WLINE:
			GSsetcolor(gr->color);
			GSPline(wx+gr->x, wy+gr->y,
				wx+gr->wd, wy+gr->ht);
			if (gr->x == gr->wd)
				GSPline(wx+gr->x+1, wy+gr->y,
					wx+gr->wd+1, wy+gr->ht);
			else
				GSPline(wx+gr->x, wy+gr->y+1,
					wx+gr->wd, wy+gr->ht+1);
			break;
						
		case GR_WRECT:
		case GR_RECT:
			GSsetcolor(gr->color);
			if (gr->attrib == FILL)
				GSPfill_rect(wx+gr->x, wy+gr->y, gr->wd, gr->ht);
			else {
				GSPframe_rect(wx+gr->x, wy+gr->y, gr->wd, gr->ht);
				if (gr->type == GR_WRECT)
					GSPframe_rect(wx+gr->x+1, wy+gr->y+1, gr->wd-2, gr->ht-2);
				}
			break;

		case GR_LABEL:
		case GR_LABEL_SMALL:
			GSsetfont(gr->type==GR_LABEL? LARGE : SMALL);
			if (strcmp(g->grshadowcol, "NONE")) {
				GSsetcolor(GSconvertcolor(g->grshadowcol));
				GSPdraw_string(wx+gr->x+1, wy+gr->y+gr->ht+1, gr->text);
				}
			GSsetcolor(GSconvertcolor(g->grcol));
			GSPdraw_string(wx+gr->x, wy+gr->y+gr->ht, gr->text);
			break;
		}
}

#define SP 8

void DrawIEField(IEField *fld, int wx, int wy)
{
	/* 	Fields are drawn with black diagonal lines across their 
		interiors.
	*/
	int x = fld->x + wx, wd = fld->wd;
	int y = fld->y + wy, ht = fld->ht;
	int x1, y1, x2, y2;
	int t;
	
	GSsetlinesize(1);
	GSsetcolor(GSconvertcolor("black"));
	GSwdrawrect(x, y, fld->wd, fld->ht);
	
	for (t=SP; (t <= wd+ht-SP); t+=SP) {
		x1 = MIN(x+t, x+wd-1);
		y1 = MAX(y, y+(t-wd));
		x2 = MAX(x, x+(t-ht));
		y2 = MIN(y+t, y+ht-1);
		GSwdrawline(x1, y1, x2, y2);
		}
}

void DrawIEDragArea(IEDragArea *dr, int wx, int wy)
{
	int x = wx + dr->x, y = wy + dr->y;
	int px, py;
	
	GSsetcolor(GSconvertcolor(dr->col));
	GSwfillrect(x, y, dr->wd, dr->ht);
	if (GSstricmp(dr->col, "black") == 0)
		GSsetcolor(GSconvertcolor("white"));
	else
		GSsetcolor(GSconvertcolor("black"));
	GSsetlinesize(1);
	GSwdrawrect(x, y, dr->wd, dr->ht);
		
	for (px = x+SP/2; px < x+dr->wd; px += SP)
		for (py = y+SP/2; py < y+dr->ht; py += SP)
			GSwdrawpoint(px, py);
	
	for (px = x+SP; px < x+dr->wd; px += SP)
		for (py = y+SP; py < y+dr->ht; py += SP)
			GSwdrawpoint(px, py);
}

void DrawIEWindow(
	IEWindow *w, 
	int clipx, 
	int clipy,
	int clipwd,
	int clipht)
{
	IEGraphic *gr;
	IEField *fld;
	IEDragArea *dr;
	IEButton *bt;
	GSColor col;
	int i;
	
	if (!w->visibleineditor)
		return;
	
	PushClip(w->x, w->y, w->wd, w->ht);
	col = GSconvertcolor(w->bgcol);
	GSsetcolor(col);

	GSPfill_rect(w->x, w->y, w->wd, w->ht);
	
	GSsetlinesize(THIN);
	GSsetcolor(GSconvertcolor(w->bdrcol));
	for (i=0; i < w->bordersize; i++)
		GSPframe_rect (w->x + i,
		               w->y + i,
		               w->wd - 2 * i,
		               w->ht - 2 * i);     

	gr = w->firstgraphic;	
	while (gr != NULL) {
		if (r_intersect(gr->gr.x+w->x, gr->gr.y+w->y, 
				gr->gr.wd, gr->gr.ht,
				clipx, clipy, clipwd, clipht))
					DrawIEGraphic(gr, w->x, w->y);
		gr = gr->nextgraphic;
		}

	fld = w->firstfield;
	while (fld != NULL) {
		if (r_intersect(fld->x+w->x, fld->y+w->y, fld->wd, fld->ht,
							clipx, clipy, clipwd, clipht))
				DrawIEField(fld, w->x, w->y);
		fld = fld->nextfield;
		}
		
	dr = w->firstdragarea;
	while (dr != NULL) {
		if (r_intersect(dr->x+w->x, dr->y+w->y, dr->wd, dr->ht,
							clipx, clipy, clipwd, clipht))
				DrawIEDragArea(dr, w->x, w->y);
		dr = dr->nextdragarea;
		}
	
	bt = w->firstbutton;
	while (bt != NULL) {
		if (bt->visibleineditor &&
			r_intersect(bt->x+w->x, bt->y+w->y, bt->wd, bt->ht,
				clipx, clipy, clipwd, clipht))
					DrawIEButton(bt, w->x, w->y);
		bt = bt->nextbutton;
		}
	PopClip();	
}

void DrawIEPopup(IEPopup *p, int clipx, int clipy, int clipwd, int clipht)
{
	IEMenuItem *item = p->itemlisthead;
	int xoff, y = p->y;
	
	if (!(p->visibleineditor))
		return;
	
	GSsetcolor(GSconvertcolor(p->bg));
	GSPfill_rect(p->x, p->y, p->wd, p->ht);
	GSsetcolor(GSconvertcolor(p->bcolor));
	GSPframe_rect(p->x, p->y, p->wd, p->ht);
	GSsetfont(LARGE);
	
	while (item != NULL) {
		if (item->active != INVISIBLE) {
			xoff = POPUPOFFSET;
			if (p->checklist)
				xoff += CHECKWID;
			y += TSIZE;
			GSsetcolor(GSconvertcolor((item->active == ENABLED)? p->efg : p->dfg));
			GSPdraw_string(p->x+xoff, y, item->text);
			}
		item = item->nextmenuitem;
		}
}	

void DrawFromPopup(IEPopup *p, int clipx, int clipy, int clipwd, int clipht)
{
	while (p != NULL) {
		DrawIEPopup(p, clipx, clipy, clipwd, clipht);
		p = p->nextpopup;
		}
}

void DrawFromWind(IEWindow *w, int clipx, int clipy, int clipwd, int clipht)
{
	while (w != NULL) {
		DrawIEWindow(w, clipx, clipy, clipwd, clipht);
		w = w->nextwin;
		}
	DrawFromPopup(iestate.popuplisthead, clipx, clipy, clipwd, clipht);
}
	
void DrawInterface(boolean changed)
{
/*	
	IEWindow *w = iestate.winlisthead;*/
	
	if ((changed)&&(twindVisible))
		GSremovenamedwindow("toolw");
	
/*	was: GSsetactivenamedwindow("interfacew"); */
	GSsetcurrentnamedwindow("interfacew");
	GSsetcolor(GSconvertcolor("white"));
	GSwfillrect(0, 0, 640, 480);
	
	DrawFromWind(iestate.winlisthead, 0, 0, SCREENWIDTH, SCREENHEIGHT);
	
	if ((changed)&&(twindVisible))
		GSdrawnamedwindow("toolw", BSTORE, NO_POPUP);
}

void SwapPalette(
	char oldPalette[], char newPalette[],
	char *oldCollist[], char *newCollist[],
	char *oldAlias[], char *newAlias[],
	char *oldAColor[], char *newAColor[],
	char *o1, char *n1,
	char *o2, char *n2
	) 
{
	char t1, *t2;
	int i, j;
	
	for (i=0; i < 256; i++) {
		SWAP(oldCollist[i], newCollist[i], t2);
		for (j=0; j<3; j++) {
			/*SWAP(oldPalette[3*i+j], newPalette[3*i+j], t1);*/
			t1 = oldPalette[3*i+j];
			oldPalette[3*i+j] = newPalette[3*i+j];
			newPalette[3*i+j] = t1;
			}
		}
	for (i=0; i < MAXALIASES+1; i++) {
		SWAP(oldAlias[i], newAlias[i], t2);
		SWAP(oldAColor[i], newAColor[i], t2);
		}
	/* not a problem on Mac or X, what about DOS??? */
}

void UpdateStateButtons(int newState)
{
	int i;
	for (i=0; i<NUM_STATES; i++) {
		if (i==newState)
			GSdisablebutton("toolw", StateButtonName[i], twindVisible);
		else if (i==curState || curState == NUM_STATES)
			GSenablebutton("toolw", StateButtonName[i], twindVisible);
		}
	curState = newState;
}

void SetDefaultObjects()
{
/*	strcpy(dfltWindow.fgcol, dfltFGCol);
	strcpy(dfltWindow.bgcol, dfltBGCol);
	strcpy(dfltWindow.bdrcol, dfltBdrCol);*/
	dfltWindow.visibleineditor = TRUE;
	dfltWindow.bordersize = 2;
	dfltWindow.active = TRUE;
	dfltWindow.bstore_to_use = 0;
	dfltWindow.firstbutton = dfltWindow.lastbutton = NULL;
	dfltWindow.firstdragarea = dfltWindow.lastdragarea = NULL;
	dfltWindow.firstgraphic = dfltWindow.lastgraphic = NULL;
	dfltWindow.firstfield = dfltWindow.lastfield = NULL;
	
	/*dfltButton.text[0] = dfltButton.funcname[0] = '\0';*/
	strcpy(dfltButton.text, "text");
	strcpy(dfltButton.funcname, "funcname");
	dfltButton.border = TRUE;
	dfltButton.repeatable = FALSE;
	dfltButton.active = ENABLED;
	dfltButton.textjust = CENTER;
	dfltButton.nextbutton = dfltButton.prevbutton = NULL;
	dfltButton.tsize = LARGE;
	dfltButton.visibleineditor = TRUE;
	
	dfltPopup.visibleineditor = TRUE;
	dfltPopup.title[0] = '\0';
	dfltPopup.function[0] = '\0';
	dfltPopup.checklist = FALSE;
	dfltPopup.itemlisthead = NULL;
	dfltPopup.itemlisttail = NULL;
}

void DrawInterfaceWind()
{
	interfaceWind = TRUE;
	GSremovenamedwindow("message");
	GSremovenamedwindow("root");
	SwapPalette(Interface_Palette, (char *) IEInterface_Palette,
				Colorlist, IEColorlist,
				Alias, IEAlias,
				Acolor, IEAcolor, NULL, NULL, NULL, NULL);
	GSPinit_palette(Interface_Palette, 256);
	GSdrawnamedwindow("interfacew", NO_BSTORE, NO_POPUP);
	GSsetcurrentnamedwindow("interfacew");
	DrawInterface(FALSE);
	
	twindVisible = TRUE;
	GSdrawnamedwindow("toolw", BSTORE, NO_POPUP);
	curState = NUM_STATES;
	UpdateStateButtons(SEL_STATE);
	
	GSgetwindowspecs("toolw", &twindx, &twindy, &twindwd, &twindht, NULL);
	
	GScharfunc("ikey");
	SetDefaultObjects();
	
	dragProcessFunc = NULL;
}

extern boolean windPropPop;

void RedrawInterfaceWind()
{
	GSredrawnamedwindow("interfacew");
	DrawInterface(FALSE);
	if (twindVisible)
		GSredrawnamedwindow("toolw");
}

void RemoveInterfaceWind()
{
	if (twindVisible)
		GSremovenamedwindow("toolw");
	
	twindVisible = FALSE;
	interfaceWind = FALSE;
	
	SwapPalette(Interface_Palette, (char *) IEInterface_Palette,
				IEColorlist, Colorlist,
				IEAlias, Alias,
				IEAcolor, Acolor, NULL, NULL, NULL, NULL);
	GSPinit_palette(Interface_Palette, 256);
	GSremovenamedwindow("interfacew");
	GSdrawnamedwindow("root", NO_BSTORE, NO_POPUP);
	GSdrawnamedwindow("message", NO_BSTORE, NO_POPUP);
	GSdisplayversion("version.txt");
	GScharfunc("process_key");
	DrawMiscWind();
}

void ToggleToolWind()
{
	if (twindVisible)
		GSremovenamedwindow("toolw");
	else
		GSdrawnamedwindow("toolw", BSTORE, NO_POPUP);
		
	twindVisible = !twindVisible;
}

boolean ToolWindowEvent(int x, int y, DragStatusType status)
{
	/* if drag starts in tool wind, then toolwind must handle */
	
	static boolean twindHandles = FALSE, twindDrag = FALSE;
	
	boolean inTWind = 
		point_in_rect(x, y, twindx, twindy, twindwd, twindht);
		
	if (!twindVisible)
		return FALSE;
		
	if (status==DRAG_INIT) {
		if (inTWind) {
			twindHandles = TRUE;
			if /*(point_in_rect(x-twindx, y-twindy, 
						tdragx, tdragy, tdragwd, tdraght))*/
						(TRUE) {
				twindDrag = TRUE;
				StartToolDrag(x, y);
				}
			return TRUE;
			}
		else {
			twindHandles = FALSE;
			return FALSE;
			}
		}
		
	else if (status==DRAG_PROCESS) {
		if (twindDrag)
			DragItem(x, y, status);
		return twindHandles;
		}
		
	else if (status==DRAG_FINAL) {
		boolean temp = twindHandles;
		if (twindDrag)
			EndToolDrag();
		twindHandles = twindDrag = FALSE;
		return temp;
		}

	else if (status==MOUSE_CLICK) {
		twindHandles = FALSE;
		return inTWind;
		}
		
	return FALSE;
}

boolean idrag(int x, int y, DragStatusType status)
{
/*	"tools" window gets first shot at handling event */

	if (ToolWindowEvent(x, y, status))
		return TRUE;

	switch (status) {
			
		case DRAG_INIT:
			if (curState == SEL_STATE)
				SelectObjectHitBy(x, y, status);
			else
				StartCreateObject(x, y, status);
			break;
			
		case DRAG_PROCESS:
/*			if (curState == SEL_STATE)
				DragItem(x, y, status);
			else
				GrowItem(x, y, status);*/
/*			if (StateFunc[curState].drag_process_func)
				(*StateFunc[curState].drag_process_func)(x, y, status);*/
			if (dragProcessFunc)
				(*dragProcessFunc)(x, y, status);
			break;
			
		case DRAG_FINAL:
			if (curState == SEL_STATE)
				ReleaseItem(x, y, status);
			else
				EndCreateObject(x, y, status);
			break;
	
		case MOUSE_INSIDE:
		case MOUSE_OUTSIDE:
			break;
			
		case MOUSE_CLICK:
			if (curState == SEL_STATE)
				ClickObject(x, y, status);
			else {
				StartCreateObject(x, y, DRAG_INIT);
				EndCreateObject(x, y, DRAG_FINAL);
				}
			break;
			
		default:
			break;
		}
	
	return TRUE;

}

boolean ikey(int x, int y, char ch)
{
#ifdef TAB_FOR_SS
	if (ch == '\t')
		dump_screen(" ");
#endif
	ToggleToolWind();
	return TRUE;
}

boolean show_winds()
{
	DrawVWindsPopup();
}

boolean set_vis_state(char *bname)
{
	int i = GSiselword(bname, StateButtonName);

	if (i == -1)
		GSelog(1, "Button name |%s| not found.\n", bname);
	
	UpdateStateButtons(i);
}

boolean vis_save()
{
	SwapPalette(Interface_Palette, (char *) IEInterface_Palette,
				IEColorlist, Colorlist,
				IEAlias, Alias,
				IEAcolor, Acolor, NULL, NULL, NULL, NULL);
	save_interface(); /* +++ return value could be used */
	SwapPalette(Interface_Palette, (char *) IEInterface_Palette,
				Colorlist, IEColorlist,
				Alias, IEAlias,
				Acolor, IEAcolor, NULL, NULL, NULL, NULL);
}
